(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

ifdef ML_INTERFACE then
open Ml_types

ifdef ML_INTERFACE then
module Translations = struct
  module StringHash = Hashtbl.Make(
    struct
      type t = string
      let equal = (=)
      let hash = Hashtbl.hash
    end )
  
  type t = cduce_t StringHash.t
  (* Create an empty table. *)
  let create = StringHash.create
  (* Add a new translation in the table. *)
  let add tbl id value = StringHash.add tbl id value
  (* Get translation associated with [id]. *)
  let find tbl id = StringHash.find tbl id
end

ifdef ML_INTERFACE then
let translations = Translations.create 17

(***********************************************************************************)

ifdef ML_INTERFACE then
let cd_type_nil = 
  let nil = Atoms.atom ( Atoms.V.mk_ascii "nil" ) in
  Types.atom nil

ifdef ML_INTERFACE then
let rec from_ocaml_rec left array = function
  | { ml_id = id; ml_kind = ML_type; ml_decl = decl } ->
      Types.descr ( from_ocaml_type left array id decl )
  | { ml_kind = ML_value; ml_decl = decl } -> 
      from_decl left array decl
  | { ml_kind = ML_module _ } -> assert false

and from_ocaml_type left array id decl = 
  try
    Translations.find translations id
  with Not_found ->
    let node = Types.make () in
    Translations.add translations id node;
    let desc = from_decl left array decl in
    Types.define node desc;
    node

and from_decl left array = function
  | ML_abstract desc -> 
      Types.descr ( from_desc left array desc )
  | ML_record list ->
      let list = List.map (
	fun ( name, desc ) ->
	  let name_ns = Ns.mk_ascii "" in
	  let name_val = Encodings.Utf8.mk name in
	  let desc = from_desc left array desc in
	  ( name_ns, name_val ), Types.descr desc
      ) list in
      Types.rec_of_list ~opened:left list
  | ML_variant list ->
      List.fold_left (
	fun t ( name, list ) ->
	  let node = Types.make () in
	  let atom = Atoms.atom ( Atoms.V.mk_ascii name ) in
	  Types.define node ( Types.atom atom );
	  let desc = match list with
	    | [] -> 
		Types.descr node
	    | hd :: [] ->
		let desc = from_desc left array hd in
		Types.times node desc
	    | _  ->
		let tuple = variant_of_list left array list in
		Types.times node tuple 
	  in
	  Types.cup t desc
      ) Types.empty list
  | ML_nil -> assert false

and from_desc left array = function
  | ML_arrow ( lbl, d1, d2 ) ->
      let node = Types.make () in
      let l = from_desc true array d1
      and r = from_desc false array d2 in
      Types.define node ( Types.arrow l r );
      node
  | ML_bool ->
      let node = Types.make () in
      Types.define node ( Builtin_defs.bool );
      node
  | ML_char ->
      let node = Types.make () in
      Types.define node ( Types.char ( Chars.mk_classes [ ( 0, 255 ) ] ) );
      node
  | ML_ident ( id, params, ocaml ) ->
      from_ocaml_type left params ocaml.ml_id ocaml.ml_decl
  | ML_int -> 
      let node = Types.make () in
      let lB = Intervals.V.mk ( string_of_int min_int )
      and uB = Intervals.V.mk ( string_of_int max_int ) in
      Types.define node ( Types.interval ( Intervals.bounded lB uB ) );
      node
  | ML_list d ->
      let desc = from_desc left array d in
      Sequence.star_node desc
  | ML_option desc ->
      let node = Types.make () in
      let trans = Types.descr ( from_desc left array desc ) in
      Types.define node ( Types.cup trans cd_type_nil );
      node
  | ML_reference desc ->
      let node = Types.make () in
      let get = ML_arrow ( None, ML_unit, desc )
      and set = ML_arrow ( None, desc, ML_unit ) in
      let ref = ML_record [ "get", get ; "set", set ] in
      Types.define node ( from_decl left array ref );
      node
  | ML_string ->
      let node = Types.make () in
      Types.define node ( Builtin_defs.string_latin1 );
      node
  | ML_tuple list ->
      assert ( List.length list >= 2 );
      tuple_of_list left array list
  | ML_unit -> 
      let node = Types.make () in
      Types.define node Sequence.nil_type;
      node
  | ML_var id -> 
      from_desc left array array.( id )

and tuple_of_list left array list = 
  let rec internal tail = function 
    | hd :: [] -> 
	let node = Types.make () in
	let desc = from_desc left array hd in
	Types.define node ( Types.times desc tail );
	node
    | hd :: tl ->
	let node = Types.make () in
	let desc = from_desc left array hd in
	Types.define node ( Types.times desc tail );
	internal node tl 
    | [] -> assert false in
  let tail = from_desc left array ( List.hd list ) in
  internal tail ( List.tl list )  

and variant_of_list left array list = 
  let rec internal tail = function 
    | hd :: [] -> 
	let node = Types.make () in
	let desc = from_desc left array hd in
	Types.define node ( Types.times tail desc );
	node
    | hd :: tl ->
	let node = Types.make () in
	let desc = from_desc left array hd in
	Types.define node ( Types.times tail desc );
	internal node tl 
    | [] -> assert false in
  let tail = from_desc left array ( List.hd list ) in
  internal tail ( List.tl list )

(***********************************************************************************)

ifdef ML_INTERFACE then
module Type : sig
  type t = cduce_t
  val from_ocaml : ocaml_t -> t
  val print : Format.formatter -> t -> unit
  val is_subtype : t -> t -> bool
end = struct
  type t = cduce_t
  let from_ocaml ocaml = from_ocaml_rec true [||] ocaml
  let print = Types.Print.print
  let is_subtype t1 t2 = Types.subtype t1 t2
end

(***********************************************************************************)

ifdef ML_INTERFACE then
module CompUnit : sig
  type t
  val from_types_cu : string -> Types.CompUnit.t -> t
  val from_bytecode : string -> t
  val find_type : t -> string -> cduce_t
  val find_value : t -> string -> cduce_t
  val find_value_slot : t -> string -> int
  val module_name : t -> string
end = struct
  type t = 
      { cu_mod: string;
	cu_unit: Types.CompUnit.t; 
	cu_typer: Typer.t; }

  let from_types_cu mod_name cu = 
    let ty = !( Typer.from_comp_unit ) cu in
    { cu_mod = mod_name; cu_unit = cu; cu_typer = ty }
      
  let from_bytecode file_name =
    let mod_name = Filename.chop_suffix file_name ".cdo" in
    let cu = Types.CompUnit.mk ( Ident.U.mk_latin1 mod_name ) in
    Librarian.import cu;
    let ty = !( Typer.from_comp_unit ) cu in
    { cu_mod = mod_name; cu_unit = cu; cu_typer = ty } 
    
  let find_type cu name = 
    let id = Ident.ident ( Ident.U.mk_latin1 name ) in
    Typer.find_type id cu.cu_typer
      
  let find_value cu name = 
    let id = Ident.ident ( Ident.U.mk_latin1 name ) in
    Typer.find_value id cu.cu_typer   

  let find_value_slot cu ( name : string ) =
    let cu = cu.cu_unit in
    let env = !Compile.from_comp_unit cu in
    let id = Ident.ident ( Ident.U.mk_latin1 name ) in
    let f_slot = Compile.find id env in
    match f_slot with Lambda.Global i -> i 
      | _ -> assert false

  let module_name cu = cu.cu_mod
end;;
