(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

ifdef ML_INTERFACE then
open Ml_types

(***********************************************************************************)

ifdef ML_INTERFACE then
(* Context used to decide how to print a type. *)
module Ctx = struct
  type t = { 
    ctx_cu: Ml_cduce.CompUnit.t;
    (* Parameter counter. *)
    ctx_counter: int }
  (* Create top position. *)
  let create cu = 
    { ctx_cu = cu;
      ctx_counter = 0 } 
  (* Get CDuce compilation unit. *)
  let comp_unit ctx = ctx.ctx_cu
  (* Generate a parameter name. *)
  let generate_name ctx = 
    { ctx with ctx_counter = (ctx.ctx_counter + 1) },
    Format.sprintf "cdo2cmo__p__%d" ctx.ctx_counter
end

(***********************************************************************************)

(* ******************************************************************************* *
 * Our translation scheme is the following:                                        *
 * Consider the following value : val f : int -> int, we have to translate left    *
 * most [ int ] as an argument from ocaml value to cduce one. The right one has    *
 * to be translated as a return value from cduce to ocaml.                         *
 * In fact we'll translate this declaration by something like:                     *
 * let f cdo2cmo_0 =                                                               *
 *  let cdo2cmo_0 = (* its translation to CDuce, see CD_of_int *) in               *
 *  let cdo2cmo_r = (* load f according to its slot in CDuce comp. env. *) in      *
 *  let cdo2cmo_r = (* apply cdo2cmo_r to cdo2cmo_0, see CD_apply *) in            *
 * match cdo2cmo_r with (* translate to int, see CD_to_int *)                      *
 * ******************************************************************************* *)

ifdef ML_INTERFACE then
type code_t =
  | To_CDuce_bool
  | To_CDuce_char
  | To_CDuce_fun of string * string * code_t * code_t
  | To_CDuce_int
  | To_CDuce_list of code_t
  | To_CDuce_option of code_t
  | To_CDuce_record of ( string * code_t ) list
  | To_CDuce_string
  | To_CDuce_tuple of code_t list
  | To_CDuce_type of string
  | To_CDuce_unit
  | To_CDuce_variant of ( string * code_t list ) list
  | To_OCaml_bool
  | To_OCaml_char
  | To_OCaml_fun of string * code_t * code_t
  | To_OCaml_int
  | To_OCaml_list of code_t
  | To_OCaml_option of code_t
  | To_OCaml_record of ( string * code_t ) list
  | To_OCaml_string
  | To_OCaml_tuple of code_t list
  | To_OCaml_type of string
  | To_OCaml_unit
  | To_OCaml_variant of ( string * code_t list ) list

ifdef ML_INTERFACE then
type value_t =
  | Type_to_cd of string * bool * bool * code_t
  | Type_to_ml of string * bool * bool * code_t
  | Value of string * int * code_t

(***********************************************************************************)

ifdef ML_INTERFACE then
let rec dump fmt = function
  | Type_to_cd ( name, first, is_rec, code ) ->
      let () = match first, is_rec with
	| true, true  -> Format.fprintf fmt "let rec "
	| true, false -> Format.fprintf fmt "let "
	| false, _    -> Format.fprintf fmt "and "
      in
      Format.fprintf fmt "ocaml2cduce__%s cdo2cmo__val = @." name;
      dump_code fmt "cdo2cmo__val" code;
      Format.fprintf fmt "@."
  | Type_to_ml ( name, first, is_rec, code ) ->
      let () = match first, is_rec with
	| true, true  -> Format.fprintf fmt "let rec "
	| true, false -> Format.fprintf fmt "let "
	| false, _    -> Format.fprintf fmt "and "
      in
      Format.fprintf fmt "cduce2ocaml__%s cdo2cmo__val = @." name;
      dump_code fmt "cdo2cmo__val" code;
      Format.fprintf fmt "@."
  | Value ( name, slot, code ) ->
      Format.fprintf fmt "let %s = @." name;
      Format.fprintf fmt "let cdo2cmo__val = @.";
      Format.fprintf fmt "!CDuce_all.Eval.from_comp_unit cdo2cmo__cu %d in@." slot;
      dump_code fmt "cdo2cmo__val" code

and dump_code fmt name = function
  | To_CDuce_bool ->
      Format.fprintf fmt "CDuce_all.Value.vbool %s@." name
  | To_CDuce_char ->
      assert false (* TODO *)
  | To_CDuce_fun ( func, arg, code1, code2 ) -> 
      Format.fprintf fmt "let %s = fun %s -> @." func arg;
      Format.fprintf fmt "let cdo2cmo__p = @.";
      dump_code fmt arg code1;
      Format.fprintf fmt "in @.";
      Format.fprintf fmt "let cdo2cmo__val = %s cdo2cmo__p in@." func;
      dump_code fmt "cdo2cmo__val" code2;
      Format.fprintf fmt "in@.";
      Format.fprintf fmt "CDuce_all.Value.Abstraction( [], %s )@." func
  | To_CDuce_int ->
      Format.fprintf fmt "CDuce_all.Value.const (@.";
      Format.fprintf fmt "CDuce_all.Types.Integer (@.";
      Format.fprintf fmt "CDuce_all.Intervals.V.mk (@.";
      Format.fprintf fmt "string_of_int %s ))) @." name
  | To_CDuce_list code ->
      Format.fprintf fmt "CDuce_all.Value.sequence (@.";
      Format.fprintf fmt "List.map (fun cdo2cmo__e -> @.";
      dump_code fmt "cdo2cmo__e" code;
      Format.fprintf fmt ") %s )@." name
  | To_CDuce_option code ->
      Format.fprintf fmt "match %s with @." name;
      Format.fprintf fmt "| None -> CDuce_all.Value.nil@.";
      Format.fprintf fmt "| Some %s -> (@." name;
      dump_code fmt name code;
      Format.fprintf fmt ")@."
  | To_CDuce_record list ->
      Format.fprintf fmt "CDuce_all.Value.vrecord [@.";
      let rec print_list = function
	| [] -> ()
	| [ ( field, desc ) ] -> 
	    Format.fprintf fmt "( CDuce_all.Ns.mk_ascii \"\", @.";
	    Format.fprintf fmt "CDuce_all.Encodings.Utf8.mk \"%s\" ), @. " field;
	    dump_code fmt ( Format.sprintf "%s.%s" name field ) desc
	| ( field, desc ) :: tl ->
	    Format.fprintf fmt "( CDuce_all.Ns.mk_ascii \"\", @.";
	    Format.fprintf fmt "CDuce_all.Encodings.Utf8.mk \"%s\" ), @. " field;
	    dump_code fmt ( Format.sprintf "%s.%s;" name field ) desc;
	    print_list tl in
      print_list list;
      Format.fprintf fmt " ]@."
  | To_CDuce_string ->
      Format.fprintf fmt "CDuce_all.Value.string_latin1 %s@." name
  | To_CDuce_unit ->
      Format.fprintf fmt "CDuce_all.Value.sequence [] @."
  | To_CDuce_tuple list ->
      let rec print_proj n = function
	| [] -> assert false
	| [ x ] -> Format.fprintf fmt "%s__%d " name n
	| hd :: tl -> 
	    Format.fprintf fmt "%s__%d, " name n;
	    print_proj ( n + 1 ) tl in
      let rec print_list n = function
	| d1 :: [ d2 ] ->
	    Format.fprintf fmt "CDuce_all.Value.Pair ( @.";
	    dump_code fmt ( Format.sprintf "%s__%d" name n ) d1;
	    Format.fprintf fmt ", ";
	    dump_code fmt ( Format.sprintf "%s__%d" name ( n + 1 ) ) d2;
	    Format.fprintf fmt ")"
	| hd :: tl ->
	    Format.fprintf fmt "CDuce_all.Value.Pair ( @.";
	    dump_code fmt ( Format.sprintf "%s__%d" name n ) hd;
	    Format.fprintf fmt ", ";
	    print_list ( n + 1 ) tl;
	    Format.fprintf fmt ")"
	| _ -> assert false in
      Format.fprintf fmt "let ( ";
      print_proj 0 list;
      Format.fprintf fmt ") = %s in@." name;
      print_list 0 list
  | To_CDuce_type id ->
      Format.fprintf fmt "ocaml2cduce__%s %s@." id name
  | To_CDuce_variant list ->
      let rec print_proj n m =
	if n < m then begin
	  Format.fprintf fmt "cdo2cmo__v%d" n;
	  if ( n + 1 ) < m then Format.fprintf fmt ", ";
	  print_proj ( n + 1 ) m
	end; in
      let rec print_list n = function
	| [ cde ] ->
	    dump_code fmt "cdo2cmo__v0" cde
	| d1 :: [ d2 ] ->
	    Format.fprintf fmt "CDuce_all.Value.Pair ( @.";
	    dump_code fmt ( Format.sprintf "cdo2cmo__v%d" n ) d1;
	    Format.fprintf fmt ", ";
	    dump_code fmt ( Format.sprintf "cdo2cmo__v%d" ( n + 1 ) ) d2;
	    Format.fprintf fmt ")@."
	| d1 :: tl ->
	    Format.fprintf fmt "CDuce_all.Value.Pair ( @.";
	    dump_code fmt ( Format.sprintf "cdo2cmo__v%d" n ) d1;
	    Format.fprintf fmt ", ";
	    print_list ( n + 1 ) tl;
	    Format.fprintf fmt ")@." 
	| _ -> assert false in
      let print ( label, list ) = match list with
	| [] ->
	    Format.fprintf fmt "| %s -> @." label;
	    Format.fprintf fmt "CDuce_all.Value.Atom ( @.";
	    Format.fprintf fmt "CDuce_all.Atoms.V.mk_ascii \"%s\" )@." label;
	| list ->
	    Format.fprintf fmt "| %s ( " label;
	    print_proj 0 ( List.length list );
	    Format.fprintf fmt " ) -> @.";
	    Format.fprintf fmt "CDuce_all.Value.Pair ( @.";
	    Format.fprintf fmt "CDuce_all.Value.Atom ( @.";
	    Format.fprintf fmt "CDuce_all.Atoms.V.mk_ascii \"%s\" ),@." label;
	    print_list 0 list;
	    Format.fprintf fmt ")@." in
      Format.fprintf fmt "match %s with @." name;
      List.iter print list
  | To_OCaml_bool ->
      Format.fprintf fmt " CDuce_all.Value.equal %s CDuce_all.Value.vtrue@." name
  | To_OCaml_char ->
      assert false (* TODO *)
  | To_OCaml_fun ( lbl, code1, code2 ) ->
      Format.fprintf fmt "fun %s -> @." lbl;
      Format.fprintf fmt " let cdo2cmo__p = @.";
      dump_code fmt lbl code1;
      Format.fprintf fmt " in @.";
      Format.fprintf fmt " let cdo2cmo__val = @.";
      Format.fprintf fmt " try @.";
      Format.fprintf fmt " CDuce_all.Eval.eval_apply cdo2cmo__val cdo2cmo__p @.";
      Format.fprintf fmt " with CDuce_all.Value.CDuceExn t -> ( @.";
      Format.fprintf fmt " match t with @.";
      Format.fprintf fmt " | CDuce_all.Value.String_latin1 ( _, _, s, _ ) -> @.";
      Format.fprintf fmt "   raise ( Error s )@.";
      Format.fprintf fmt " | CDuce_all.Value.String_utf8 ( _, _, t, _ ) -> @.";
      Format.fprintf fmt "   raise ( Error ( @.";
      Format.fprintf fmt "   CDuce_all.Encodings.Utf8.get_str t ) )@.";
      Format.fprintf fmt " | _ -> assert false ) in@.";
      dump_code fmt "cdo2cmo__val" code2
  | To_OCaml_int ->
      Format.fprintf fmt " match CDuce_all.Value.inv_const %s with@." name;
      Format.fprintf fmt "  | CDuce_all.Types.Integer i -> @.";
      Format.fprintf fmt "     CDuce_all.Intervals.V.get_int i@.";
      Format.fprintf fmt "  | _ -> assert false@."
  | To_OCaml_list code -> 
      Format.fprintf fmt " let rec cdo2cmo__seq list = function @.";
      Format.fprintf fmt " | CDuce_all.Value.Atom _ -> list@.";
      Format.fprintf fmt " | CDuce_all.Value.Pair (cdo2cmo__hd, cdo2cmo__tl) ->@.";
      Format.fprintf fmt "    let cdo2cmo__hd = ( @.";
      dump_code fmt "cdo2cmo__hd" code;
      Format.fprintf fmt "   ) in cdo2cmo__seq ( list @@ [ cdo2cmo__hd ] ) @.";
      Format.fprintf fmt "    cdo2cmo__tl @.";
      Format.fprintf fmt " | CDuce_all.Value.Concat (cdo2cmo__l1, cdo2cmo__l2) ->@.";
      Format.fprintf fmt "    ( cdo2cmo__seq list cdo2cmo__l1 ) @@ @.";
      Format.fprintf fmt "    ( cdo2cmo__seq [] cdo2cmo__l2 )@.";
      Format.fprintf fmt " | _ -> assert false in@.";
      Format.fprintf fmt " cdo2cmo__seq [] %s@." name
  | To_OCaml_option code ->
      Format.fprintf fmt " match %s with@." name;
      Format.fprintf fmt " | CDuce_all.Value.nil -> None@.";
      Format.fprintf fmt " | %s -> Some (@." name;
      dump_code fmt name code;
      Format.fprintf fmt " )@."
  | To_OCaml_record list -> 
      Format.fprintf fmt " match %s with@." name;
      Format.fprintf fmt " | CDuce_all.Value.Record cdo2cmo__map ->@.";
      let rec trans = function
	| [] -> ()
	| ( field, code )::tl ->
	    Format.fprintf fmt "let cdo2cmo__%s = @." field;
	    dump_code fmt ( 
	      "( CDuce_all.Ident.LabelMap.assoc ( " ^ (
		"CDuce_all.Ident.LabelPool.mk ( CDuce_all.Ns.mk_ascii \"\"," 
	      ) ^ (
		Format.sprintf
		"CDuce_all.Encodings.Utf8.mk \"%s\" ) ) cdo2cmo__map )" field 
	      )
	    ) code;
	    Format.fprintf fmt "in@.";
	    trans tl in
      trans list;
      let rec record = function
	| [] -> ()
	| ( field, _ ) :: tl ->
	    Format.fprintf fmt "%s = cdo2cmo__%s; " field field;
	    record tl in
      Format.fprintf fmt "{ ";
      record list;
      Format.fprintf fmt " }@.";
      Format.fprintf fmt "| _ -> assert false@."
  | To_OCaml_string ->
      Format.fprintf fmt " CDuce_all.Value.get_string_latin1 %s@." name
  | To_OCaml_tuple list ->
      let rec proj n = function
	| [ _ ; _ ] -> 
	    Format.fprintf fmt "CDuce_all.Value.Pair ( @.";
	    Format.fprintf fmt "cdo2cmo__%d, cdo2cmo__%d )@." n ( n + 1 )
	| _ :: tl ->
	    Format.fprintf fmt "CDuce_all.Value.Pair ( cdo2cmo__%d,@." n;
	    proj ( n + 1 ) tl;
	    Format.fprintf fmt ")@." 
	| _ -> assert false in
      let rec print_list n = function
	| [ d1 ; d2 ] ->
	    Format.fprintf fmt "(@.";
	    dump_code fmt ( Format.sprintf "cdo2cmo__%d" n ) d1;
	    Format.fprintf fmt "), (@.";
	    dump_code fmt ( Format.sprintf "cdo2cmo__%d" ( n + 1 ) ) d2;
	    Format.fprintf fmt ")@."
	| hd :: tl ->
	    Format.fprintf fmt "(@.";
	    dump_code fmt ( Format.sprintf "cdo2cmo__%d" n ) hd;
	    Format.fprintf fmt "),@.";
	    print_list ( n + 1 ) tl 
	| _ -> assert false in
      Format.fprintf fmt "match %s with " name;
      proj 0 list;
      Format.fprintf fmt "-> (@.";
      print_list 0 list;
      Format.fprintf fmt ")@.";
      Format.fprintf fmt " | _ -> assert false@."
  | To_OCaml_type id ->
      Format.fprintf fmt "cduce2ocaml__%s %s@." id name
  | To_OCaml_unit ->
      Format.fprintf fmt " ()@."
  | To_OCaml_variant list ->
      let rec print_proj n m =
	if n < m then begin
	  Format.fprintf fmt "cdo2cmo__v%d" n;
	  if n + 1 < m then Format.fprintf fmt ",@.";
	  print_proj ( n + 1 ) m
	end; in
      let rec print_list = function
	| [] -> 
	    Format.fprintf fmt " | _ -> assert false @."
	| ( field, [] ) :: tl ->
	    Format.fprintf fmt " | \"%s\" -> %s@." field field;
	    print_list tl
	| ( field, list ) :: tl ->
	    Format.fprintf fmt " | \"%s\" -> ( @." field;
	    let cnt = ref 0 in
	    List.iter (
	      fun desc ->
		Format.fprintf fmt "let cdo2cmo__v%d = @." !cnt;
		dump_code fmt "cdo2cmo__desc" desc;
		Format.fprintf fmt "in@.";
		incr cnt
	    ) list;
	    Format.fprintf fmt "%s ( " field;
	    print_proj 0 ( List.length list );
	    Format.fprintf fmt ") ) @.";
	    print_list tl in
      Format.fprintf fmt " let cdo2cmo__atom, cdo2cmo__desc = match %s with @." name;
      Format.fprintf fmt " | CDuce_all.Value.Atom cdo2cmo__atom -> @.";
      Format.fprintf fmt "   cdo2cmo__atom, CDuce_all.Value.nil @.";
      Format.fprintf fmt " | CDuce_all.Value.Pair ( @.";
      Format.fprintf fmt "   CDuce_all.Value.Atom cdo2cmo__atom, @.";
      Format.fprintf fmt "   cdo2cmo__desc ) -> @.";
      Format.fprintf fmt "   cdo2cmo__atom, cdo2cmo__desc @.";
      Format.fprintf fmt " | _ -> assert false in@.";
      Format.fprintf fmt " match CDuce_all.Ns.QName.to_string ( @.";
      Format.fprintf fmt "   CDuce_all.Atoms.V.value cdo2cmo__atom )@.";
      Format.fprintf fmt " with @.";
      print_list list

ifdef ML_INTERFACE then
let dump_list fmt list =
  List.iter ( fun value -> dump fmt value ) list

ifdef ML_INTERFACE then
let rec generate_ml ctx name desc = 
  let slot = ml_value_slot ctx name in
  let code = generate_to_ml ctx "cdo2cmo__val" desc in
  Value ( name, slot, code )

and generate_to_ml ctx name = function
  | ML_abstract desc ->
      generate_desc_to_ml ctx name desc
  | ML_record list ->
      let list = List.map (
	fun ( field, desc ) ->
	  let d = generate_desc_to_ml_rec ctx desc
	  in field, d
      ) list in To_OCaml_record list
  | ML_variant list ->
      let list = List.map (
	fun ( field, desc ) -> match desc with
	  | [] -> 
	      field, []
	  | [ x ] ->
	      let desc = generate_desc_to_ml_rec ctx x in
	      field, [ desc ]
	  | list ->
	      let list = List.map ( generate_desc_to_ml_rec ctx ) list in
	      field, list
      ) list in 
      To_OCaml_variant list
  | ML_nil -> assert false

and generate_desc_to_ml ctx name desc =
  generate_desc_to_ml_rec ctx (*name*) desc
    
and generate_desc_to_ml_rec ctx (* name *) = function
  | ML_bool   -> To_OCaml_bool
  | ML_char   -> To_OCaml_char
  | ML_int    -> To_OCaml_int
  | ML_string -> To_OCaml_string
  | ML_unit   -> To_OCaml_unit
  | ML_arrow ( lbl, d1, d2 ) ->
      let ctx, lbl = generate_ml_arg ctx lbl in
      let code1 = generate_desc_to_cd ctx lbl d1 in
      let code2 = generate_desc_to_ml ctx lbl d2 in
      To_OCaml_fun ( lbl, code1, code2 )
  | ML_ident ( _, _, { Ml_types.ml_id = id } ) -> 
      To_OCaml_type id
  | ML_list d -> 
      To_OCaml_list ( generate_desc_to_ml_rec ctx d )
  | ML_option d ->
      To_OCaml_option ( generate_desc_to_ml_rec ctx d )
  | ML_reference _ -> assert false
  | ML_tuple list ->
      let list = List.map ( generate_desc_to_ml_rec ctx ) list in
      To_OCaml_tuple list
  | ML_var _ -> assert false

and generate_to_cd ctx name = function   
  | ML_abstract desc ->
      generate_desc_to_cd ctx name desc
  | ML_record list -> 
      let list = List.map (
	fun ( field, desc ) ->
	  let d = generate_desc_to_cd_rec ctx name desc in
	  field, d
      ) list in To_CDuce_record list
  | ML_variant list ->
      let list = List.map (
	fun ( field, desc ) -> match desc with
	  | [] -> 
	      field, []
	  | [ x ] ->
	      let desc = generate_desc_to_cd_rec ctx name x in
	      field, [ desc ]
	  | _ ->
	      let list = List.map ( generate_desc_to_cd_rec ctx name ) desc in
	      field, list
      ) list in 
      To_CDuce_variant list
  | ML_nil -> assert false

and generate_desc_to_cd ctx name desc = 
  generate_desc_to_cd_rec ctx name desc

and generate_desc_to_cd_rec ctx name = function
  | ML_bool   -> To_CDuce_bool
  | ML_char   -> To_CDuce_char 
  | ML_int    -> To_CDuce_int
  | ML_string -> To_CDuce_string
  | ML_unit   -> To_CDuce_unit
  | ML_arrow ( lbl, d1, d2 ) -> 
      let ctx, lbl = Ctx.generate_name ctx in
      let code1 = generate_desc_to_ml ctx lbl d1 in
      let code2 = generate_desc_to_cd ctx name d2 in
      To_CDuce_fun ( name, lbl, code1, code2 )
  | ML_ident ( _, _, { Ml_types.ml_id = id } ) -> 
      To_CDuce_type id
  | ML_list d -> 
      To_CDuce_list ( generate_desc_to_cd_rec ctx name d )
  | ML_option d -> 
      To_CDuce_option ( generate_desc_to_cd_rec ctx name d )
  | ML_reference d -> assert false
      (* To_CDuce_reference ( generate_desc_to_cd_rec ctx name d ) *)
  | ML_tuple list ->
      let list = List.map ( generate_desc_to_cd_rec ctx name ) list in
      To_CDuce_tuple list
  | ML_var _ -> assert false

and generate_ml_arg ctx = function
  | Some lbl -> ctx, lbl
  | None -> Ctx.generate_name ctx 

and ml_value_slot ctx = Ml_cduce.CompUnit.find_value_slot ( Ctx.comp_unit ctx )

ifdef ML_INTERFACE then
let generate_type_to_ml ctx is_rec list =
  let rec internal first tail = function
    | [] -> tail
    | { Ml_types.ml_id = id; Ml_types.ml_rec = r; Ml_types.ml_decl = decl } :: tl ->
	let is_rec = is_rec || r in
	let code = generate_to_ml ctx "cdo2cmo__val" decl in
	internal false ( tail @ [ Type_to_ml ( id, first, is_rec, code ) ] ) tl
  in
  internal true [] list

ifdef ML_INTERFACE then
let generate_type_to_cd ctx is_rec list =
  let rec internal first tail = function
    | [] -> tail
    | { Ml_types.ml_id = id; Ml_types.ml_rec = r; Ml_types.ml_decl = decl } :: tl ->
	let is_rec = r || is_rec in
	let code = generate_to_cd ctx "cdo2cmo__val" decl in
	internal false ( tail @ [ Type_to_cd ( id, first, is_rec, code ) ] ) tl
  in
  internal true [] list
      
(***********************************************************************************)

ifdef ML_INTERFACE then
let ml_print fmt cu t =
  let name = t.ml_name
  and decl = t.ml_decl
  and kind = t.ml_kind in
  let code = generate_ml ( Ctx.create cu ) name decl in dump fmt code

(***********************************************************************************)

ifdef ML_INTERFACE then
module ML : sig
  val generate : Format.formatter -> Ml_ocaml.CompUnit.t -> Ml_cduce.CompUnit.t -> unit
end = struct
  let generate fmt ml_cu cd_cu =
    let mod_name = Ml_cduce.CompUnit.module_name cd_cu in
    Format.fprintf fmt "exception Error of string@.";
    Format.fprintf fmt "let cdo2cmo__cu = @.";
    Format.fprintf fmt " let cdo2cmo__cu = @.";
    Format.fprintf fmt " CDuce_all.Types.CompUnit.mk (@.";
    Format.fprintf fmt "  CDuce_all.Ident.U.mk_latin1 \"%s\") in@." mod_name;
    Format.fprintf fmt " CDuce_all.Librarian.import cdo2cmo__cu;@.";
    Format.fprintf fmt " CDuce_all.Librarian.run (CDuce_all.Value.sequence []) ";
    Format.fprintf fmt "cdo2cmo__cu;@.";
    Format.fprintf fmt "cdo2cmo__cu@.";
    Ml_ocaml.CompUnit.iter ( 
      fun ( need, list ) -> match list with
	| [ { Ml_types.ml_kind = Ml_types.ML_type } as ml_type ] ->
	    Ml_ocaml.Type.print fmt ( need, [ ml_type ] );
	    let code = generate_type_to_ml ( Ctx.create cd_cu ) false [ ml_type ] in
	    dump_list fmt code;
	    let code = generate_type_to_cd ( Ctx.create cd_cu ) false [ ml_type ] in
	    dump_list fmt code
	| ( { Ml_types.ml_kind = Ml_types.ML_type } :: tl ) ->
	    Ml_ocaml.Type.print fmt ( need, list );
	    let code = generate_type_to_ml ( Ctx.create cd_cu ) true list in
	    dump_list fmt code;
	    let code = generate_type_to_cd ( Ctx.create cd_cu ) true list in
	    dump_list fmt code
	| [ { Ml_types.ml_kind = Ml_types.ML_value } as ml_type ] ->
	    ml_print fmt cd_cu ml_type
	| _ -> ()
    ) ml_cu
end
