(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

ifdef ML_INTERFACE then
open Ml_types

(***********************************************************************************)

ifdef ML_INTERFACE then
type error =
  | File_not_found of string
  | Unsupported_feature of string

ifdef ML_INTERFACE then
exception Error of error

ifdef ML_INTERFACE then
let error e = raise ( Error e )

ifdef ML_INTERFACE then
let report_error err =
  let msg = match err with
    | File_not_found s ->
	"ml interface : error file " ^ s ^ " not found in object path"
    | Unsupported_feature s -> 
	"ml interface : error unsupported " ^ s ^ " in interface file (.mli)"
  in
  Format.eprintf "%s@." msg

(***********************************************************************************)

ifdef ML_INTERFACE then
module ML = OCaml_all

(***********************************************************************************)

ifdef ML_INTERFACE then
let recursive_type = ref false

(***********************************************************************************)

ifdef ML_INTERFACE then
let rec ml_print_list fmt separator f = function
  | [] -> assert false
  | hd :: [] -> f fmt hd
  | hd :: tl -> 
      f fmt hd; 
      Format.fprintf fmt "%s" separator;
      ml_print_list fmt separator f tl

ifdef ML_INTERFACE then
let rec ml_print fmt is_rec name decl = function
  | ML_type ->
      let () = if is_rec then Format.fprintf fmt "and %s = " name
      else Format.fprintf fmt "type %s = " name in
      ml_print_decl fmt decl;
      Format.fprintf fmt "@."
  | ML_value ->
      Format.fprintf fmt "val %s : " name;
      ml_print_decl fmt decl;
      Format.fprintf fmt "@."
  | ML_module list -> 
      Format.fprintf fmt "module %s : sig @." name;
      List.iter ( 
	fun t -> ml_print fmt false t.ml_name t.ml_decl t.ml_kind 
      ) list;
      Format.fprintf fmt "end@."

and ml_print_decl fmt = function
  | ML_abstract desc -> ml_print_desc fmt desc
  | ML_record list -> 
      Format.fprintf fmt "{ ";
      ml_print_list fmt "; " ( 
	fun fmt ( name, desc ) -> 
	  Format.fprintf fmt "%s: " name;
	  ml_print_desc fmt desc 
      ) list;
      Format.fprintf fmt " }"
  | ML_variant list -> 
      ml_print_list fmt " | " ( 
	fun fmt ( name, list ) -> match list with
	  | [] -> Format.fprintf fmt "%s" name
	  | hd :: [] -> 
	      Format.fprintf fmt "%s of " name;
	      ml_print_desc fmt hd
	  | list ->
	      Format.fprintf fmt "%s of " name;
	      ml_print_list fmt " * " ml_print_desc list
      ) list
  | _ -> assert false

and ml_print_desc fmt = function
  | ML_arrow ( None, d1, d2 ) ->
      Format.fprintf fmt "( ";
      ml_print_desc fmt d1;
      Format.fprintf fmt " ) -> ";
      ml_print_desc fmt d2
  | ML_arrow ( Some lbl, d1, d2 ) ->
      Format.fprintf fmt "%s:( " lbl;
      ml_print_desc fmt d1;
      Format.fprintf fmt " ) -> ";
      ml_print_desc fmt d2
  | ML_bool -> Format.fprintf fmt "bool" 
  | ML_char -> Format.fprintf fmt "char"
  | ML_ident ( name, [||], _ ) -> 
      Format.fprintf fmt "%s" name
  | ML_ident ( name, array, _ ) ->
      Format.fprintf fmt "( ";
      let le = ( Array.length array ) - 1 in
      for i = 0 to le do
	ml_print_desc fmt array.( i );
	if i < le then Format.fprintf fmt ","
      done;
      Format.fprintf fmt " ) %s" name
  | ML_int -> Format.fprintf fmt "int"
  | ML_list desc -> 
      ml_print_desc fmt desc;
      Format.fprintf fmt " list"
  | ML_option desc -> 
      ml_print_desc fmt desc;
      Format.fprintf fmt " option"
  | ML_reference desc ->
      ml_print_desc fmt desc;
      Format.fprintf fmt " ref"
  | ML_string -> Format.fprintf fmt "string"
  | ML_tuple list -> 
      Format.fprintf fmt "(";
      ml_print_list fmt " * " ml_print_desc list;
      Format.fprintf fmt ")"
  | ML_unit -> Format.fprintf fmt "unit"
  | ML_var id -> Format.fprintf fmt "(%d)" id

(***********************************************************************************)

ifdef ML_INTERFACE then
module Type : sig
  type t = ocaml_t
  val print : Format.formatter -> ( bool * t list )-> unit
end = struct
  type t = ocaml_t;;
  let print fmt ( need, list ) =
    let rec internal is_rec = function
      | [ t ] -> 
	  ml_print fmt is_rec t.ml_name t.ml_decl t.ml_kind
      | t::tl ->
	  ml_print fmt is_rec t.ml_name t.ml_decl t.ml_kind;
	  internal true tl
      | _ -> ()
    in
    if need then internal false list else ()
end

(***********************************************************************************)

ifdef ML_INTERFACE then
module Path = struct
  type t = string list * string  
  (* [append path name] append [name] to the end of [path]. *)
  let append (list,last) name = (list @ [ last ], name)
  (* [create list name] create a path from a [list] and a [name]. *)
  let create list name = list, name
  (* [from_string string] create a path from [string]. *)
  let from_string string =
    let rec from_string_rec : 
      string list -> string -> string list * string option =
      fun tail str ->
	try
	  let idx = String.index str '.' in
	  let tmp = String.sub str 0 idx in
	  let idx = idx + 1 in
	  let res = String.sub str idx ((String.length str) - idx) in
	  let p,o = from_string_rec (tail @ [ tmp ]) res in
	  p, None
	with Not_found -> tail, Some str
    in
    let path,name = from_string_rec [] string in
    let name = match name with Some s -> s | None -> assert false in
    path, name
  (* Get path head. *)
  let head ( list, last ) = List.hd list
  (* Get path length. *)
  let length ( list, last ) = 1 + ( List.length list )
  (* [prepend path name] prepend [name] to the beginning of [path]. *)
  let prepend ( list,last ) name = ( name :: list, last )
  (* Remove path queue. *)
  let remove_end ( list, last ) =
    let rec internal tail = function
      | [ x ] -> tail, x
      | hd :: tl -> internal ( tail @ [ hd ] ) tl 
      | _ -> assert false in
    let list, path = internal [] list in
    list, path
  (* [to_string path] translate [path] to string. *)
  let to_string ( path, name ) =
    let sub = 
      List.fold_left (fun prev str -> prev ^ str ^ "_") "" path in
    sub ^ name
end

(***********************************************************************************)

ifdef ML_INTERFACE then
module Env = struct
  (* Set of forward references. *)
  module StringSet = Set.Make ( String )
  (* Map of local types. *)
  module StringMap = Map.Make ( String )
  (* Main type. *)
  type t = 
      { (* Current module path. *)
	env_path: Path.t;
	(* List of values. *)
	env_list: ( bool * ocaml_t list ) list;
	(* Type queue to merge recursive types. *)
	env_queue: ocaml_t list;
	(* Local environment. *)
	env_local: string StringMap.t;
	(* Global table containing all types. *)
	env_global: ( string, ocaml_t ) Hashtbl.t;
        (* Forward references. *)
	env_forward: StringSet.t;
        (* Are we dealing with external module. *)
	env_external: bool;
        (* Parameters table. *)
	env_parameters: ( int, int ) Hashtbl.t }
  (* Create an empty environment. *)
  let current name = 
    { env_path = Path.from_string name;
      env_list = [];
      env_queue = [];
      env_local = StringMap.empty;
      env_global = Hashtbl.create 17;
      env_forward = StringSet.empty;
      env_external = false;
      env_parameters = Hashtbl.create 17 }
  (* Create an external environment. *)
  let external_env env name =
    { env with
      env_path = Path.from_string name;
      env_list = [];
      env_queue = [];
      env_local = StringMap.empty;
      env_forward = StringSet.empty;
      env_external = true;
      env_parameters = Hashtbl.create 17 }
  (* Add a pair to the environment. *)
  let rec add env = function
    | { ml_id = id; ml_name = name; ml_kind = ML_type } as value ->
	{ env with
	  env_queue = env.env_queue @ [ value ];
	  env_local = StringMap.add name id env.env_local }
    | { ml_name = name; ml_kind = ML_value } as value ->
	let len = Path.length env.env_path = 1 in 
	let list = env.env_list @ [ len, [ value ] ] in
	{ env with env_list = list }
    | value -> 
	let len = Path.length env.env_path = 1 in 
	let list = env.env_list @ [ len, [ value ] ] in
	{ env with env_list = list }
  (* Add a forward reference. *)
  let add_forward_ref env name = 
    { env with env_forward = StringSet.add name env.env_forward }
  (* Add a new parameter mapping. *)
  let add_parameter env id cnt = Hashtbl.add env.env_parameters id cnt
  (* Get current module path. *)
  let current_path env = env.env_path
  (* Get type in local environment. *)
  let find_local env key = try
    StringMap.find key env.env_local
  with Not_found ->
    Path.to_string ( Path.append env.env_path key )
  (* Get parameter value. *)
  let find_parameter env id = try
    Hashtbl.find env.env_parameters id
  with Not_found -> id
  (* Get type in global environment. *)
  let find_type env id name = try
    Hashtbl.find env.env_global id
  with Not_found -> 
    let node = 
      { ml_id = id;
	ml_name = name;
	ml_kind = ML_type;
	ml_rec = false;
	ml_decl = ML_nil } in
    Hashtbl.add env.env_global node.ml_id node;
    node
  (* Flush recursive type buffer. *)
  let flush_types env = match env.env_queue with
    | [] -> env
    | values ->
	let len = Path.length env.env_path = 1 in
	{ env with 
	  env_list = env.env_list @ [ len, values ];
	  env_queue = [] }
  (* Is there any forward reference. *)
  let has_forward_ref env = not ( StringSet.is_empty env.env_forward )
  (* Are we treating local file. *)
  let is_file_local env = not env.env_external
  (* Is there a type in global environment called ... *)
  let mem_type env = Hashtbl.mem env.env_global
  (* Add a module to module path. *)
  let push_module env name = 
    { env with env_path = Path.append env.env_path name }
  (* Pop top module. *)
  let pop_module env =
    { env with env_path = Path.remove_end env.env_path }
  (* Remove forward reference. *)
  let remove_forward_ref env name = 
    { env with env_forward = StringSet.remove name env.env_forward }
  (* Get an environment without parameters. *)
  let without_parameters env = { env with env_parameters = Hashtbl.create 17 }
end

(***********************************************************************************)

ifdef ML_INTERFACE then
(* Translate internal OCaml compiler Path.t to our Path.t *)
let path_to_mlpath path =
  let rec internal_rec tail = function
    | ML.Path.Pident id     -> (ML.Ident.name id) :: tail
    | ML.Path.Pdot (p,id,_) -> internal_rec (id::tail) p
    | _ -> failwith "path_to_mlpath : Papply -- (2)" in
  match path with
    | ML.Path.Pident id     -> [], ML.Ident.name id
    | ML.Path.Pdot (p,id,_) -> internal_rec [] p, id
    | _ -> failwith "path_to_mlpath : Papply -- (1)"

(***********************************************************************************)

ifdef ML_INTERFACE then
let rec translate_ident env path spath = function
  | [] when spath = "bool" -> env, ML_bool
  | [] when spath = "char" -> env, ML_char
  | [] when spath = "int" -> env, ML_int
  | [] when spath = "string" -> env, ML_string
  | [] when spath = "unit" -> env, ML_unit
  | [] -> 
      let env, node = match path with 
	| ML.Path.Pident _ ->
	    let env, node = 
	      let lpath = Env.find_local env spath in
	      match Env.find_type env lpath spath with
	      | { ml_decl = ML_nil } as node ->
		  Env.add_forward_ref env spath, node
	      | node -> env, node in
	    env, node
	| ML.Path.Pdot _ ->
	    let mde = ML.Ident.name ( ML.Path.head path ) in
	    if not ( Env.mem_type env spath ) then load_module env mde;
	    let node = Env.find_type env spath spath in
	    env, node
	| _ -> assert false
      in
      env, ML_ident ( spath, [||], node )
  | [ hd ] when spath = "list" ->
      let env, desc = translate_texpr env hd in
      env, ML_list desc
  | [ hd ] when spath = "option" ->
      let env, desc = translate_texpr env hd in
      env, ML_option desc
  | [ hd ] when spath = "array" ->
      error ( Unsupported_feature "array" )
  | [ hd ] when spath = "Pervasives.ref" ->
      let env, desc = translate_texpr env hd in
      env, ML_reference desc
  | list ->
      let env, prms = List.fold_left (
	fun ( env, list ) texpr ->
	  let env, desc = translate_texpr env texpr in
	  env, list @ [ desc ]
      ) ( env, [] ) list in
      let mde = ML.Ident.name ( ML.Path.head path ) in
      let name = Env.find_local env spath in
      if not ( Env.mem_type env name ) then load_module env mde;
      let node = Env.find_type env spath spath in
      env, ML_ident ( spath, Array.of_list prms, node )

and load_module env name =
  let file = String.copy name in
  String.set file 0 ( Char.lowercase ( String.get file 0 ) );
  let file = find_file ( Format.sprintf "%s.cmi" file ) in
  let sign = ML.Env.read_signature name file in
  let _ = translate_signature ( Env.external_env env name ) sign in ()

and find_file name =
  let dir =
    try
      List.find ( 
	fun dir -> Sys.file_exists ( Filename.concat dir name ) 
      ) !Librarian.obj_path
    with Not_found -> error ( File_not_found name )
  in
  if dir = "" then error ( File_not_found name );
  Filename.concat dir name

and translate_tdesc env id = function
  | ML.Types.Tvar -> 
      env, ML_var id
  | ML.Types.Tarrow ( lbl, e1, e2, commut ) -> 
      let (str : string) = Obj.magic lbl in
      let str = if str = "" then None else Some str in
      let _, te1 = translate_texpr env e1
      and _, te2 = translate_texpr env e2 in
      env, ML_arrow ( str, te1, te2 )
  | ML.Types.Ttuple ( list ) -> 
      let mapping e = let _,e = translate_texpr env e in e in 
      env, ML_tuple ( List.map mapping list )
  | ML.Types.Tconstr ( path, list, abbrev_ref ) ->
      let p = path_to_mlpath path in
      let p = Path.to_string p in
      translate_ident env path ( ML.Path.name path ) list
  | ML.Types.Tvariant _ ->
      error ( Unsupported_feature "polymorphic variant" );
  | _ -> 
      failwith "translate_tdesc : any"

and translate_texpr env { ML.Types.id = id; ML.Types.desc = tdesc } =
  let id = Env.find_parameter env id in
  translate_tdesc env id tdesc
    
and translate_variants env list = 
  let internal ( env, tail ) ( name, list ) = match list with
    | [] -> 
	env, tail @ [ name, [] ]
    | hd :: [] ->
	let env, desc = translate_texpr env hd in
	env, tail @ [ name, [ desc ] ]
    | list -> 
	let env, list = List.fold_left (
	  fun ( env, tail ) desc ->
	    let env, desc = translate_texpr env desc in
	    env, tail @ [ desc ]
	) ( env, [] ) list in
	env, tail @ [ name, list ]
  in
  List.fold_left internal ( env, [] ) list
    
and translate_fields env list =
  List.map (
    fun (name, _, texpr) -> 
      let _,e = translate_texpr env texpr in name, e
  ) list
   
and translate_tdecl env tdecl =
  let { ML.Types.type_params = params;
	ML.Types.type_arity = arity; 
	ML.Types.type_kind = kind; 
	ML.Types.type_manifest = manifest } = tdecl in
  let is_nok  = arity <> 0 && Env.is_file_local env in 
  if is_nok then error (Unsupported_feature "polymorphic type");
  let cnt = ref 0 in
  List.iter ( 
    fun { ML.Types.id = id } -> 
      Env.add_parameter env id !cnt; incr cnt
  ) params;
  match kind with
    | ML.Types.Type_abstract when manifest = None -> 
	error (Unsupported_feature "abstract type")
    | ML.Types.Type_abstract -> 
	let texpr = match manifest with Some e -> e | _ -> assert false in 
	let env, desc = translate_texpr env texpr in
	env, ML_abstract desc
    | ML.Types.Type_variant ( list, priv ) -> 
	let env, list = translate_variants env list in
	env, ML_variant list
    | ML.Types.Type_record ( list, _, _ ) ->
	env, ML_record ( translate_fields env list )

and translate_mdecl env id = function
  | ML.Types.Tmty_signature s ->
      let id = ML.Ident.name id in
      let env, list = translate_signature ( Env.push_module env id ) s in
      let ocaml = {
	ml_id = id;
	ml_rec = false;
	ml_name = id;
	ml_kind = ML_module list;
	ml_decl = ML_nil; } in
      let env =  Env.pop_module env in
      Env.add env ocaml, ocaml
  | _ -> failwith "translate_mdecl : any"

and translate_signature_item env = function
  | ML.Types.Tsig_type ( id, tdecl ) -> 
      let name =  ML.Ident.name id in
      let path = Env.current_path env in
      let id = ( Path.to_string path ) ^ "_" ^ name in
      let env = Env.without_parameters env in
      let env, decl = translate_tdecl env tdecl in
      let ocaml = Env.find_type env id name in
      ocaml.ml_rec <- Env.has_forward_ref env;
      ocaml.ml_decl <- decl;
      let env = Env.remove_forward_ref env id in
      let env = if not ocaml.ml_rec then Env.flush_types env else env in
      Env.add env ocaml, Some ocaml
  | ML.Types.Tsig_value ( id, { ML.Types.val_type = texpr } ) -> 
      let name = ML.Ident.name id in
      let path = Env.current_path env in
      let id = ( Path.to_string path ) ^ "_" ^ name in
      let env, desc = translate_texpr env texpr in
      let ml_val = { 
	ml_id = id;
	ml_rec = false;
	ml_name = name; 
	ml_kind = ML_value;
	ml_decl = ML_abstract desc; } in
      Env.add ( Env.flush_types env ) ml_val, Some ml_val
  | ML.Types.Tsig_exception ( id, list ) -> 
      env, None
  | ML.Types.Tsig_module ( id, mdecl )  ->
      let env, t_ml = translate_mdecl env id mdecl in
      env, Some t_ml
  | ML.Types.Tsig_modtype ( id, ( ML.Types.Tmodtype_manifest m ) ) ->
      let env, t_ml = translate_mdecl env id m in
      env, Some t_ml
  | ML.Types.Tsig_modtype ( id, ( ML.Types.Tmodtype_abstract ) ) ->
      error (Unsupported_feature "abstract module")
  | ML.Types.Tsig_cltype _ 
  | ML.Types.Tsig_class _ -> 
      error (Unsupported_feature "class")

and translate_signature env signature =
  List.fold_left (
    fun (e,list) s -> 
      let e, l = translate_signature_item e s in 
      let list = match l with Some l -> list @ [ l ] | _ -> list in
      e, list 
  ) ( env, [] ) signature
      
(***********************************************************************************)

ifdef ML_INTERFACE then
module CompUnit : sig 
  type t
  val from_bytecode : string -> string -> t
  val iter : ( ( bool * Type.t list ) -> unit ) -> t -> unit
end = struct
  type t = { cu_list: ( bool * ocaml_t list ) list }

  let from_bytecode file_name mod_name =
    String.set mod_name 0 (Char.uppercase (String.get mod_name 0));
    let signature = ML.Env.read_signature mod_name file_name in
    let env = Env.current mod_name in
    try
      let env, list = translate_signature env signature in
      { cu_list = env.Env.env_list }
    with
      | Error err -> report_error err; exit 1

  let iter f cu = List.iter f cu.cu_list
end
