(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

open Location
open Ident

module C = Types.CompUnit

exception InconsistentCrc of C.t
exception Loop of C.t
exception InvalidObject of string
exception CannotOpen of string
exception NoImplementation of C.t

type t = {
  typing: Typer.t;
  compile: Compile.env;
  code: Lambda.code_item list;
  mutable digest: Digest.t option;
  mutable vals: Value.t array option;
  mutable depends: C.t list
}

let mk (typing,compile,code) =
  { typing = typing;
    compile = compile;
    code = code;
    digest = None;
    vals = None;
    depends = [];
  }

let magic = "CDUCE:compunit:00001"

let obj_path = ref [ "" ]

let tbl = C.Tbl.create ()

let find id =
  try C.Tbl.find tbl id
  with Not_found -> assert false

let serialize s cu =
  Serialize.Put.magic s magic;
  Typer.serialize s cu.typing;
  Compile.serialize s cu.compile;
  Lambda.Put.codes s cu.code

let deserialize s =
  Serialize.Get.magic s magic;
  let typing = Typer.deserialize s in
  let compile = Compile.deserialize s in
  let code = Lambda.Get.codes s in
  mk (typing,compile,code)

let serialize_dep=
  Serialize.Put.list 
    (Serialize.Put.pair Encodings.Utf8.serialize Serialize.Put.string)

let deserialize_dep =
  Serialize.Get.list
    (Serialize.Get.pair Encodings.Utf8.deserialize Serialize.Get.string)


let find_obj id = 
  let base = Encodings.Utf8.to_string (C.value id) ^ ".cdo" in
  let p = 
    List.find (fun p -> Sys.file_exists (Filename.concat p base)) !obj_path in
  Filename.concat p base

let save id out =
  protect_op "Save compilation unit";
  Lambda.Put.current_cu := id;
  let cu = find id in
  C.enter id;
  let raw = Serialize.Put.run serialize cu in
  let depend = C.close_serialize () in
  C.leave ();

(*
  print_endline "Dependencies:";
  List.iter (fun x -> print_endline (object_filename x)) depend;
  flush stdout;
*)
  let depend = 
    try List.map 
      (fun id ->
	 match (C.Tbl.find tbl id).digest with
	   | Some d -> (C.value id, d)
	   | None -> assert false
      ) depend
    with Not_found -> assert false in

  let depend = Serialize.Put.run serialize_dep depend in
  let digest = Digest.string raw in
  let oc = open_out out in
  Marshal.to_channel oc (digest,depend,raw) [];
  close_out oc
  
  
let check_digest id exp digest =
  match digest with
    | Some x ->
	if exp <> x then 
	  raise (InconsistentCrc id)
    | None -> 
	assert false

let loop = C.Tbl.create ()
let check_loop id = 
  try 
    C.Tbl.find loop id;
    raise (Loop id)
  with Not_found -> 
    C.Tbl.add loop id ()

let depends = ref []
let during_compile = ref false

let show ppf id t v =
  match id with
    | Some id ->
	Format.fprintf ppf "@[val %a : @[%a@]@."
	U.print (Id.value id)
	Types.Print.print t 
    | None -> ()

let rec compile verbose id src =
  check_loop id;
  protect_op "Compile external file";
  let ic = 
    try open_in src
    with Sys_error _ -> raise (CannotOpen src) in
  Location.push_source (`File src);
  let input = Stream.of_channel ic in
  let p = 
    try Parser.prog input 
    with
    | Stdpp.Exc_located (_, (Location _ as e)) -> raise e
    | Stdpp.Exc_located ((i,j), e) -> raise_loc i j e
  in
  close_in ic;
  let argv = ident (U.mk "argv") in
  during_compile := true;
  C.enter id;
  let show =
    if verbose 
    then Some (show Format.std_formatter)
    else None in
  let cu =
    Compile.comp_unit 
      ?show
      (Typer.enter_value argv (Sequence.star Sequence.string) Builtin.env)
      (Compile.enter_global Compile.empty argv)
      p
  in
  let cu = mk cu in
  C.Tbl.add tbl id cu;
  C.leave ();
  during_compile := false;
  cu.depends <- !depends;
  depends := []

let rec load id =
  protect_op "Load compiled compilation unit";
  try 
    C.Tbl.find tbl id
  with Not_found ->
(*    Printf.eprintf "load %s: start\n" (object_filename id);
    flush stderr; *)

    let obj = 
      try find_obj id
      with Not_found -> raise (NoImplementation id) in
    let ic = 
      try open_in obj
      with Sys_error _ -> raise (CannotOpen obj) in

    let (dig, depend, raw) = 
      try Marshal.from_channel ic
      with Failure _ | End_of_file -> raise (InvalidObject obj) in

    close_in ic;
    let depend = Serialize.Get.run deserialize_dep depend in
    check_loop id;
    if !during_compile then depends := id :: !depends;
    List.iter (fun (id,dig) -> load_check (C.mk id) dig) depend;
    C.enter id;
    let cu = Serialize.Get.run deserialize raw in
    C.leave ();
    cu.depends <- List.map (fun (id,_) -> C.mk id) depend;
    cu.digest <- Some dig;
    C.Tbl.add tbl id cu;
    Typer.register_types id cu.typing;
    cu

and load_check id exp =
  let cu = load id in
  check_digest id exp cu.digest

let rec run argv id =
  let cu = find id in
  match cu.vals with
    | None -> 
	List.iter (run argv) cu.depends;
	let vals = Eval.comp_unit [argv] cu.code in
(*
	Compile.dump Format.std_formatter cu.compile;
	Array.iter (fun v ->
		      Format.fprintf Format.std_formatter "%a@."
		      Value.print v) vals;
*)
	cu.vals <- Some vals
    | Some _ -> ()
  
let import id = ignore (load id)

let () =
  Typer.from_comp_unit := (fun cu -> (load cu).typing);
  Compile.from_comp_unit := (fun cu -> (load cu).compile);
  Eval.from_comp_unit := 
  (fun cu i ->
     match (load cu).vals with
       | None -> !Eval.stack.(i) (* TODO: check that cu is being evaluated *)
       | Some a -> a.(i))
