(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

open Ident

let () = State.close ();;

let load_dump = ref None
let save_dump = ref None

let out_dir  = ref [] (* directory of the output file *)
let src  = ref []
let args = ref []

let compile = ref false
let run = ref false

ifdef ML_INTERFACE then
  let ocaml_support = true
else 
  let ocaml_support = false;;

let version () =
  Printf.eprintf "CDuce, version %s\n" <:symbol<cduce_version>>;
  Printf.eprintf "built on %s\n" <:symbol<build_date>>;
  Printf.eprintf "using OCaml %s compiler\n" <:symbol<ocaml_compiler>>;
  Printf.eprintf "support for ocaml interfacing: %b\n" ocaml_support;
  Printf.eprintf "support for expat: %b\n" (Load_xml.expat_support);
  Printf.eprintf "support for curl: %b\n" (Url.curl_support);
  Printf.eprintf "support for netclient: %b\n" (Url.netclient_support);
  exit 0

let license () =
  Printf.eprintf "\n\
The CDuce interpreter is distributed under the terms of the Q Public \n\
License version 1.0 (included in the sources). The Choice of Law section\n\
been modified from the original Q Public.\n\n
"; 
  exit 0

let specs = ref 
  [ "--load", Arg.String (fun s -> load_dump := Some s), 
          "   load persistency file before running CDuce program";
    "--save", Arg.String (fun s -> save_dump := Some s), 
          "   save persistency file after running CDuce program";
    "--dump", Arg.String (fun s -> save_dump := Some s; load_dump := Some s), 
          "   specify persistency file for loading and saving";
    "--verbose", Arg.Set Cduce.verbose,
             "verbose output for compilation (show types of exported values)";
    "--compile", Arg.Set compile,
             "compile the given CDuce file";
    "--obj-dir",  Arg.String (fun s -> out_dir := s :: !out_dir),
             "directory for the compiled .cdo file";
    "-I", Arg.String (fun s -> Librarian.obj_path := s::!Librarian.obj_path),
      "       add one directory to the lookup path for .cdo files";
    "--run", Arg.Set run,
           "    execute the given .cdo file";
    "--stdin", Arg.Unit (fun () -> src := "" :: !src),
           "  read CDuce script on standard input";
    "--debug", Arg.Unit (fun () -> Stats.set_verbosity Stats.Summary),
           "  print profiling/debugging information";
    "-v", Arg.Unit version,
      "       print CDuce version";
    "--version", Arg.Unit version,
             "print CDuce version";
    "--license", Arg.Unit license,
             "print CDuce license";
   "--arg", Arg.Rest (fun s -> args := s :: !args),
         "    following arguments are passed to the CDuce program (in argv)";  

 ]

let ppf = Format.std_formatter
let ppf_err = Format.err_formatter

let () =
  if Load_xml.expat_support then
     specs := 
    ("--expat", Arg.Unit (fun () -> Load_xml.use_parser := `Expat),
             "  use expat parser (default)") ::
    ("--pxp", Arg.Unit (fun () -> Load_xml.use_parser := `Pxp),
           "    use PXP parser") ::
    !specs
  else
    specs :=
    ("--expat", Arg.Unit (fun () -> (Format.fprintf ppf  "WARNING: --expat unused option. CDuce compiled without expat support\n\n")),
             "  option not available: CDuce was compiled without expat support") ::
    ("--pxp", Arg.Unit (fun () -> (Format.fprintf ppf  "WARNING: --pxp useless option. CDuce compiled without expat support\n\n")),
           "    useless option:  CDuce was compiled without expat support") ::
    !specs


let err s =
  prerr_endline s;
  exit 1

let mode () =
  Arg.parse !specs (fun s -> src := s :: !src) 
    "Usage:\ncduce [OPTIONS ...] [FILE ...] [--arg argument ...]\n\nOptions:";
  match (!compile,!out_dir,!run,!src,!args) with
    | false, _::_, _,  _, _   -> 
	err "--obj-dir option can be used only with --compile"
    | false, [], false, [],  args   -> `Toplevel args
    | false, [], false, [x], args   -> `Script (x,args)
    | false, [], false, _, _        ->
	err "Only one CDuce program can be executed at a time"
    | true,  [o], false, [x], []     -> `Compile (x,Some o) 
    | true,  [], false, [x], []     -> `Compile (x,None) 
    | true,  [], false, [], []      ->
	err "Please specify the CDuce program to be compiled"
    | true,  [], false, _, []       ->
	err "Only one CDuce program can be compiled at a time"
    | true,  _, false, _, []        ->
	err "Please specify only one output directory"
    | true,  _, false, _, _        ->
	err "No argument can be passed to programs at compile time"
    | false, _, true,  [x], args   -> `Run (x,args)
    | false, _, true,  [], _       ->
	err "Please specifiy the CDuce program to be executed"
    | false, _, true,   _, _       ->
	err "Only one CDuce program can be executed at a time"
    | true, _, true,   _,  _       ->
	err "The options --compile and --run are incompatible"
	


let bol = ref true

let outflush s =
  output_string stdout s;
  flush stdout

let toploop () =
  let restore = 
    try 
      let tcio = Unix.tcgetattr Unix.stdin in
      Unix.tcsetattr 
	Unix.stdin Unix.TCSADRAIN { tcio with Unix.c_vquit = '\004' };
      fun () -> Unix.tcsetattr Unix.stdin Unix.TCSADRAIN tcio
    with Unix.Unix_error (_,_,_) -> 
      fun () -> ()
  in
  let quit () = 
    outflush "\n";
    restore ();
    exit 0
  in
  Format.fprintf ppf "        CDuce version %s\n@." <:symbol<cduce_version>>;
  Sys.set_signal Sys.sigquit (Sys.Signal_handle (fun _ -> quit ()));
  Sys.catch_break true;
  Cduce.toplevel := true;
  Location.push_source `Stream;
  let read i =
    if !bol then 
      if !Ulexer.in_comment then outflush "* " else outflush "> ";
    try 
      let c = input_char stdin in
      bol := c = '\n';
      Some c
    with Sys.Break -> quit () 
  in
  let input = Stream.from read in
  let rec loop () =
    outflush "# ";
    bol := false;
    ignore (Cduce.topinput ppf ppf_err input);
    while (input_char stdin != '\n') do () done;
    loop () in
  (try loop () with End_of_file -> ());
  restore ()

let argv args = 
  Value.sequence (List.rev_map Value.string_latin1 args)

let restore argv =
  match !load_dump with
    | Some f ->
	(try 
	   Format.fprintf ppf "Restoring state: ";
	   let chan = open_in_bin f in
	   let s = Marshal.from_channel chan in
	   close_in chan;
	   State.set s; 
	   Format.fprintf ppf "done ...@."
	 with Sys_error _ ->  
	   Format.fprintf ppf "failed ...@.")
    | None -> 
	let t = Sequence.star Sequence.string in
	Cduce.enter_global_value (ident (U.mk "argv")) argv t

let save () =
  match !save_dump with
    | Some f ->
	Format.fprintf ppf "Saving state ...@\n";
	let s = State.get () in
	let chan = open_out_bin f in
	Marshal.to_channel chan s [ Marshal.Closures ];
	close_out chan
    | None -> ()

let main () = 
  match mode () with
    | `Toplevel args ->
	restore (argv args);
	toploop ();
	save ()
    | `Script (f,args) ->
	Cduce.compile_run f (argv args)
    | `Compile (f,o) ->
	Cduce.compile f o
    | `Run (f,args) ->
	Cduce.run f (argv args)
	 
let () = 
(* Hum... *)
  let b = ref true in
  at_exit (fun () -> if !b then (b := false; main ()));
  at_exit (fun () -> Stats.dump Format.std_formatter)


