(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

open Value
open Ident
open Patterns.Compile
open Encodings


type path = Left of path | Right of path | Label of label * path | Root

let rec print_path ppf = function
  | Left p -> print_path ppf p; Format.fprintf ppf "L"
  | Right p -> print_path ppf p; Format.fprintf ppf "R"
  | Label (l,p) -> print_path ppf p; Format.fprintf ppf "(lab:%s)" (Label.to_string (LabelPool.value l))
  | Root -> Format.fprintf ppf "*"
  

exception Path of path

let make_result pt fail (code,_) =
  if fail == code then raise (Path pt);
  code

let rec run_disp_basic pt fail f =  function
  | [(_,r)] -> make_result pt fail r
  | (t,r)::rem -> 
      if f t then make_result pt fail r 
      else run_disp_basic pt fail f rem
  | _ -> assert false

let find_array pred a = 
  let res = ref (-1) in 
  for i = 0 to Array.length a - 1 do
    if pred a.(i) then (assert (!res = (-1)); res := i)
  done;
  !res

let new_fail_res fail =
  find_array (function (code,_) when code = fail -> true | _ -> false)
let new_fail_disp fail =
  find_array (function Ignore (code,_) when code = fail -> true | _ -> false)


let rec run_dispatcher pt fail d v = 
  match actions d with
    | AIgnore r -> make_result pt fail r
    | AKind k -> run_disp_kind pt fail k v

and run_disp_kind pt fail actions = function
  | Pair (v1,v2) -> run_disp_prod pt fail v1 v2 actions.prod
  | Xml (v1,v2,v3) -> run_disp_prod pt fail v1 (Pair(v2,v3)) actions.xml
  | Record r -> run_disp_record pt fail false (LabelMap.get r) actions.record
  | Atom a -> make_result pt fail (Atoms.get_map a actions.atoms) 
  | Char c -> make_result pt fail (Chars.get_map c actions.chars) 
  | Integer i ->
      run_disp_basic pt fail (fun t -> Types.Int.has_int t i) actions.basic
  | Abstraction (iface,_)
  | Abstraction2 (_,iface,_) ->
      run_disp_basic pt fail (fun t -> Types.Arrow.check_iface iface t) 
        actions.basic
  | Absent ->
      run_disp_basic pt fail (fun t -> Types.Record.has_absent t) actions.basic
  | Delayed _ -> assert false
  | v ->
      run_disp_kind pt fail actions (normalize v)


and run_disp_prod pt fail v1 v2 = function
  | Impossible -> assert false
  | TailCall d1 -> run_dispatcher (Left pt) fail d1 v1
  | Ignore d2 -> run_disp_prod2 pt fail v2 d2
  | Dispatch (d1,b1) ->
      let code1 = run_dispatcher (Left pt) (new_fail_disp fail b1) d1 v1 in
      run_disp_prod2 pt fail v2 b1.(code1)

and run_disp_prod2 pt fail v2 = function
  | Impossible -> assert false
  | Ignore r -> make_result pt fail r
  | TailCall d2 -> run_dispatcher (Right pt) fail d2 v2
  | Dispatch (d2,b2) ->
      let code2 = run_dispatcher (Right pt) (new_fail_res fail b2) d2 v2 in
      make_result pt fail b2.(code2)

and run_disp_record pt fail other fields = function
  | None -> assert false
  | Some (RecLabel (l,d)) ->
      let rec aux other = function
	| (l1,_) :: rem when l1 < l -> aux true rem
	| (l1,vl) :: rem when l1 == l ->
	    run_disp_record1 pt fail l other vl rem d
	| rem -> 
	    run_disp_record1 pt fail l other Absent rem d
      in
      aux other fields
  | Some (RecNolabel (some,none)) ->
      let other = other || (fields != []) in
      let r = if other then some else none in
      match r with
	| Some r -> make_result pt fail r
	| None -> assert false
      
and run_disp_record1 pt fail l other v1 rem = function
  | Impossible -> assert false
  | TailCall d1 -> run_dispatcher (Label (l,pt)) fail d1 v1
  | Ignore d2 ->  run_disp_record2 pt fail other rem d2
  | Dispatch (d1,b1) ->
      let code1 = run_dispatcher pt (new_fail_disp fail b1) d1 v1 in
      run_disp_record2 pt fail other rem b1.(code1)

and run_disp_record2 pt fail other rem = function
  | Impossible -> assert false
  | Ignore r -> make_result pt fail r
  | TailCall d2 -> run_disp_record_loop pt fail other rem d2
  | Dispatch (d2,b2) ->
      let code2 = run_disp_record_loop pt (new_fail_res fail b2) other rem d2 in
      make_result pt fail b2.(code2)

and run_disp_record_loop pt fail other rem d =
  match actions d with
    | AIgnore r -> make_result pt fail r
    | AKind k -> run_disp_record pt fail other rem k.record
  
let explain t v =
  let p = Patterns.make IdSet.empty in
  Patterns.define p (Patterns.constr t);
  let (d,rhs) = make_branches Types.any [ (p,()) ] in
  let fail = find_array (function Fail -> true | _ -> false) rhs in
  try 
    ignore (run_dispatcher Root fail d v);
    None
  with Path p -> Some p

